USE benchmark;

DELIMITER $$
DROP PROCEDURE IF EXISTS R9$$
CREATE PROCEDURE `R9`(
  IN start BIGINT UNSIGNED,
  IN end BIGINT UNSIGNED
)
READS SQL DATA
BEGIN
  DECLARE labelmax VARCHAR(40);
  DECLARE maxcount BIGINT UNSIGNED DEFAULT 0;
  DECLARE count BIGINT UNSIGNED;
  DECLARE lab VARCHAR(40);
  DECLARE ana_id INT UNSIGNED;
  DECLARE th1 FLOAT;
  DECLARE th2 FLOAT;
  DECLARE no_more_rows BOOLEAN;
  DECLARE curana CURSOR FOR SELECT label, id, threshold_low, threshold_high FROM ana_desc;
  DECLARE CONTINUE HANDLER FOR NOT FOUND
    SET no_more_rows = TRUE;
  OPEN curana;
  the_loop: LOOP
    FETCH curana INTO lab, ana_id, th1, th2;
    IF no_more_rows THEN
      CLOSE curana;
      LEAVE the_loop;
    END IF;
    SELECT count(*) INTO count FROM ana_val WHERE id = ana_id AND (value>th2 OR value<th1) AND date BETWEEN start AND end;
    IF count > maxcount THEN
      SET maxcount = count;
      SET labelmax = lab;
    END IF;
  END LOOP the_loop;
  SELECT labelmax, maxcount;
END$$

DROP PROCEDURE IF EXISTS subR10$$
CREATE PROCEDURE `subR10`(
  IN ana_id INT UNSIGNED,
  IN start BIGINT UNSIGNED,
  IN end BIGINT UNSIGNED,
  IN period INT UNSIGNED,
  OUT count BIGINT UNSIGNED
)
READS SQL DATA
BEGIN
  DECLARE dat BIGINT UNSIGNED;
  DECLARE olddate BIGINT UNSIGNED;
  DECLARE no_more_rows BOOLEAN DEFAULT FALSE;
  DECLARE curanav CURSOR FOR SELECT date FROM ana_val WHERE id = ana_id AND date BETWEEN start AND end ORDER BY date ASC;
  DECLARE CONTINUE HANDLER FOR NOT FOUND
    SET no_more_rows = TRUE;
  OPEN curanav;
  SET count = 0;
  FETCH curanav INTO olddate;
  the_loop: LOOP
    FETCH curanav INTO dat;
    IF no_more_rows THEN
      CLOSE curanav;
      LEAVE the_loop;
    END IF;
    IF (dat - olddate) > period THEN
      SET count = count + 1;
    END IF;
    SET olddate = dat;
  END LOOP the_loop;
END$$

DROP PROCEDURE IF EXISTS R10$$
CREATE PROCEDURE `R10`(
  IN start BIGINT UNSIGNED,
  IN end BIGINT UNSIGNED
)
READS SQL DATA
BEGIN
  DECLARE idmax INT UNSIGNED;
  DECLARE maxcount BIGINT UNSIGNED DEFAULT 0;
  DECLARE per INT UNSIGNED;
  DECLARE count BIGINT UNSIGNED;
  DECLARE ana_id INT UNSIGNED;
  DECLARE no_more_rows BOOLEAN;
  DECLARE curana CURSOR FOR SELECT id, intervl FROM ana_desc;
  DECLARE CONTINUE HANDLER FOR NOT FOUND
    SET no_more_rows = TRUE;
  OPEN curana;
  the_loop: LOOP
    FETCH curana INTO ana_id, per;
    IF no_more_rows THEN
      CLOSE curana;
      LEAVE the_loop;
    END IF;
    SET per = per * 1000;
    CALL subR10(ana_id, start, end, per, count);
    IF count > maxcount THEN
      SET maxcount = count;
      SET idmax = ana_id;
    END IF;
  END LOOP the_loop;
  SELECT idmax, maxcount;
END$$
DELIMITER ;
