#include "datagen.h"
#include <stddef.h>

#include "config.h"

using namespace std;

datagen::datagen(int seed) {
  this->seed = seed;
  random = new CRandomMother(seed);
  prevtime = CFG_STARTDATE;
  anavalues_count = 0;
  boolvalues_count = 0;
}

datagen::~datagen() {
  if(random != NULL) {
    delete random;
    random = NULL;
  }
}

void datagen::reset() {
  if(random != NULL) {
    delete random;
    random = new CRandomMother(seed);
  }
  prevtime = CFG_STARTDATE;
  anavalues_count = 0;
  boolvalues_count = 0;
}

unsigned int datagen::getAnaId() {
  return random->IRandom(1,CFG_ANA_TAGS);
}

unsigned int datagen::getBoolId() {
  return random->IRandom(1,CFG_TOR_TAGS);
}

unsigned int datagen::getUnit() {
  return random->IRandom(1,CFG_UNIT);
}

unsigned int datagen::getInterval() {
  return random->IRandom(1,CFG_PERIOD);
}

float datagen::getAnaValue() {
  unsigned int val = random->BRandom();
  float * pt = (float *) &val;

  if((val & 0x7F800000) == 0x7F800000) {//NaN or inf
    val = (val & 0xFF7FFFFF);
  }

  float ret = *pt;
  return ret;
}

float datagen::getThreshold1() {
  unsigned int val = (random->BRandom() | 0x80000000); //negative sign
  float * pt = (float *) &val;

  if((val & 0x7F800000) == 0x7F800000) {//NaN or inf
    val = (val & 0xFF7FFFFF);
  }

  return *pt;
}

float datagen::getThreshold2() {
  unsigned int val = (random->BRandom() & 0x7FFFFFFF); //positive sign
  float * pt = (float *) &val;

  if((val & 0x7F800000) == 0x7F800000) {//NaN or inf
    val = (val & 0xFF7FFFFF);
  }

  return *pt;
}

bool datagen::getBoolValue() {
  return (random->BRandom() & 0x00000001);
}

unsigned long long datagen::getDate() {
  return prevtime+=random->IRandom(CFG_TSINCR_MIN, CFG_TSINCR_MAX);
}

unsigned int datagen::getCreationDate() {
  return random->IRandom(CFG_CREATEDATE/1000, CFG_STARTDATE/1000);
}

unsigned int datagen::getDestructionDate() {
  return 0;
}

void datagen::getLabel(char * label) {
  int length = random->IRandom(1,40);
  int j;
  for(j=0; j<length; j++) {
    label[j] = (char) random->IRandom(32,124);
    if(label[j]==(char)(34)) {
      label[j] = (char)125;
    }
    if(label[j]==(char)(92)) {
      label[j] = (char)126;
    }
  }
  label[j] = '\0';
}

void datagen::getThresholdLabel(char * label) {
  int length = random->IRandom(1,60);
  int j;
  for(j=0; j<length; j++) {
    label[j] = (char) random->IRandom(32,124);
    if(label[j]==(char)(34)) {
      label[j] = (char)125;
    }
    if(label[j]==(char)(92)) {
      label[j] = (char)126;
    }
  }
  label[j] = '\0';
}

unsigned char datagen::getQuality() {
  return (unsigned char) (random->IRandom(0,255));
}

void datagen::getStartEnd(long long &start, long long &end) {
  start = random->LRandom(CFG_STARTDATE, prevtime - CFG_SELECTRANGE - 1);
  end = start + CFG_SELECTRANGE;
}
