#include "dbms.h"
#include "config.h"

#include <iostream>
#include <sstream>
#include <cstdlib>
#include <math.h>
#include <string.h>

#include <chr_chronos.h>

#define CHR_DEVICE "/tmp/flash.dummy"
#define CHR_FOLDER "/tmp/"

#define ID(key) (*((uint32_t*)((unsigned char*)(key))))
#define DATE(key) (*((uint64_t*)(((unsigned char*)(key))+sizeof(uint32_t))))
#define VALUEA(buf) (*((float*)((unsigned char*)(buf))))
#define QUALITYA(buf) (*(((unsigned char*)(buf))+sizeof(float)))
#define VALUEB(buf) (*((bool*)((unsigned char*)(buf))))
#define QUALITYB(buf) (*(((unsigned char*)(buf))+sizeof(bool)))

using namespace std;

static chr_chronos * chronos;
static chr_table_ctx * anavalues;
static chr_table_ctx * boolvalues;
static chr_cursor * anacursors[CFG_ANA_TAGS];
static chr_cursor * boolcursors[CFG_TOR_TAGS];
static unsigned char * key_buf;
static unsigned char * key_end;
static unsigned char * anavalue_buf;
static unsigned char * boolvalue_buf;

const char * dbms::name()
{
  return "Chronos";
}

const char * dbms::version()
{
  return "1.0.0";
}

const char * dbms::filesystem()
{
  return "raw";
}

void dbms::print_options()
{

}

dbms::dbms()
{
  chronos = new chr_chronos(CHR_DEVICE, CHR_FOLDER);
  if(!CFG_SELECT_ONLY)
  {
    chronos->drop_table("ana_val");
    chronos->create_table("ana_val", sizeof(uint32_t) + sizeof(uint64_t),
        sizeof(float) + sizeof(char));
    chronos->drop_table("bool_val");
    chronos->create_table("bool_val", sizeof(uint32_t) + sizeof(uint64_t),
        sizeof(bool) + sizeof(char));
  }
  anavalues = chronos->open_table("ana_val");
  boolvalues = anavalues;
  boolvalues = chronos->open_table("bool_val");
  key_buf = (unsigned char *)malloc(sizeof(uint32_t)+sizeof(uint64_t));
  key_end = (unsigned char *)malloc(sizeof(uint32_t)+sizeof(uint64_t));
  anavalue_buf = (unsigned char *)malloc(sizeof(float)+sizeof(char));
  boolvalue_buf = (unsigned char *)malloc(sizeof(bool)+sizeof(char));
  if(!CFG_SELECT_ONLY)
  {
    DATE(key_buf) = 0;
    for(int i=0; i < CFG_ANA_TAGS; i++)
    {
      ID(key_buf) = htobe32(i + 1);
      anacursors[i] = anavalues->open_cursor(key_buf);
    }
    for(int i=0; i < CFG_TOR_TAGS; i++)
    {
      ID(key_buf) = htobe32(i + 1);
      boolcursors[i] = boolvalues->open_cursor(key_buf);
    }
  }
}

dbms::~dbms()
{
  chronos->close_table(anavalues);
  chronos->close_table(boolvalues);
  free(key_buf);
  free(key_end);
  free(anavalue_buf);
  free(boolvalue_buf);
  delete chronos;
}

typedef struct ana_st
{
  char * label;
  unsigned int period;
  float threshold1;
  float threshold2;
} ana_t;

static ana_t ana_tbl[CFG_ANA_TAGS];

void dbms::insertAna(uint32_t anaid, char * label, unsigned int creationdate, unsigned int * destructiondate, unsigned int unit, unsigned int period, float threshold1, float threshold2)
{
  ana_tbl[anaid - 1].label = (char *)malloc(strlen(label) + 1);
  strcpy(ana_tbl[anaid - 1].label, label);
  ana_tbl[anaid - 1].period = period;
  ana_tbl[anaid - 1].threshold1 = threshold1;
  ana_tbl[anaid - 1].threshold2 = threshold2;
}

void dbms::insertBool(uint32_t boolid, char * label, unsigned int creationdate, unsigned int * destructiondate, char * label0, char * label1)
{}

void dbms::flushR0_1()
{
  if(!CFG_SELECT_ONLY)
  {
    for(int i=0; i < CFG_ANA_TAGS; i++)
    {
      anavalues->close_cursor(anacursors[i]);
    }
  }
}

void dbms::flushR0_2()
{
  if(!CFG_SELECT_ONLY)
  {
    for(int i=0; i < CFG_TOR_TAGS; i++)
    {
      boolvalues->close_cursor(boolcursors[i]);
    }
  }
}

void dbms::queryR0_1(uint32_t anaid, uint64_t date, float ana_value, unsigned char quality)
{
  if(!CFG_SELECT_ONLY)
  {
    ID(key_buf) = htobe32(anaid);
    DATE(key_buf) = htobe64(date);
    VALUEA(anavalue_buf) = ana_value;
    QUALITYA(anavalue_buf) = quality;

    int ret = anacursors[anaid - 1]->insert(key_buf, anavalue_buf);
    if(ret != CHR_OK)
    {
      cerr << "insert returned " << ret << endl;
      exit(-1);
    }
  }
}

void dbms::queryR0_2(uint32_t boolid, uint64_t date, bool bool_value, unsigned char quality)
{
  if(!CFG_SELECT_ONLY)
  {
    ID(key_buf) = htobe32(boolid);
    DATE(key_buf) = htobe64(date);
    VALUEB(boolvalue_buf) = bool_value;
    QUALITYB(boolvalue_buf) = quality;

    int ret = boolcursors[boolid - 1]->insert(key_buf, boolvalue_buf);
    if(ret == CHR_OOR)
    {
      boolvalues->close_cursor(boolcursors[boolid - 1]);
      boolcursors[boolid - 1] = boolvalues->open_cursor(key_buf);
      ret = boolcursors[boolid - 1]->insert(key_buf, boolvalue_buf);
      if(ret != 0)
      {
        cerr << "second insert with new cursor returned " << ret << endl;
        exit(-1);
      }
    }
  }
}

void dbms::queryR1_1(float Value, uint32_t AnaId, uint64_t Date)
{
  if(!CFG_SELECT_ONLY)
  {
    ID(key_buf) = htobe32(AnaId);
    DATE(key_buf) = htobe64(Date);

    chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
    if(cur == NULL)
    {
      cerr << "open_rdcursor error" << endl;
      exit(-1);
    }

    cur->read_next(key_buf, anavalue_buf);
    if(be32toh(ID(key_buf)) != AnaId ||
        be64toh(DATE(key_buf)) != Date)
    {
      cerr << "update has wrong key (" << AnaId << ", " << Date
        << ")" << endl;
      exit(-1);
    }

    VALUEA(anavalue_buf) = Value;
    QUALITYA(anavalue_buf) = QUALITYA(anavalue_buf) | 128;
    if(cur->update(key_buf, anavalue_buf) != 0)
    {
      cerr << "update failed" << endl;
      exit(-1);
    }
    anavalues->close_cursor(cur);
  }
}

void dbms::queryR1_2(bool Value, uint32_t BoolId,
    uint64_t Date)
{
  if(!CFG_SELECT_ONLY)
  {
    ID(key_buf) = htobe32(BoolId);
    DATE(key_buf) = htobe64(Date);

    chr_cursor * cur = boolvalues->open_cursor(key_buf, CHR_KEY);
    if(cur == NULL)
    {
      cerr << "open_rdcursor error" << endl;
      exit(-1);
    }

    cur->read_next(key_buf, boolvalue_buf);
    if(be32toh(ID(key_buf)) != BoolId ||
        be64toh(DATE(key_buf)) != Date)
    {
      cerr << "update has wrong key (" << BoolId << ", " << Date
        << ")" << endl;
      exit(-1);
    }

    VALUEB(boolvalue_buf) = Value;
    QUALITYB(boolvalue_buf) = QUALITYB(boolvalue_buf) | 128;
    if(cur->update(key_buf, boolvalue_buf) != 0)
    {
      cerr << "update failed" << endl;
      exit(-1);
    }
    boolvalues->close_cursor(cur);
  }
}

void dbms::queryR2_1(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)
    (uint32_t, uint64_t, float, unsigned char))
{
  ID(key_buf) = htobe32(AnaId);
  DATE(key_buf) = htobe64(start);

  chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
  if(cur == NULL)
  {
    cerr << "open_rdcursor error" << endl;
    exit(-1);
  }

  while(cur->read_next(key_buf, anavalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == AnaId &&
      be64toh(DATE(key_buf)) < end)
  {
    callback(be32toh(ID(key_buf)), be64toh(DATE(key_buf)),
        VALUEA(anavalue_buf), QUALITYA(anavalue_buf));
  }

  anavalues->close_cursor(cur);
}

void dbms::queryR2_2(uint32_t BoolId, uint64_t start,
    uint64_t end, void (*callback)
    (uint32_t, uint64_t, bool, unsigned char))
{
  ID(key_buf) = htobe32(BoolId);
  DATE(key_buf) = htobe64(start);

  chr_cursor * cur = boolvalues->open_cursor(key_buf, CHR_KEY);
  if(cur == NULL)
  {
    cerr << "open_rdcursor error" << endl;
    exit(-1);
  }

  while(cur->read_next(key_buf, boolvalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == BoolId &&
      be64toh(DATE(key_buf)) < end)
  {
    callback(be32toh(ID(key_buf)), be64toh(DATE(key_buf)),
        VALUEB(boolvalue_buf), QUALITYB(boolvalue_buf));
  }

  boolvalues->close_cursor(cur);
}

void dbms::queryR3_1(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(uint64_t))
{
  ID(key_buf) = htobe32(AnaId);
  DATE(key_buf) = htobe64(start);

  uint64_t count = 0;

  chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
  if(cur == NULL)
  {
    cerr << "open_rdcursor error" << endl;
    exit(-1);
  }

  while(cur->read_next(key_buf, anavalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == AnaId &&
      be64toh(DATE(key_buf)) < end)
  {
    count++;
  }

  anavalues->close_cursor(cur);
  callback(count);
}

void dbms::queryR3_2(uint32_t BoolId, uint64_t start,
    uint64_t end, void (*callback)(uint64_t))
{
  ID(key_buf) = htobe32(BoolId);
  DATE(key_buf) = htobe64(start);

  uint64_t count = 0;

  chr_cursor * cur = boolvalues->open_cursor(key_buf, CHR_KEY);
  if(cur == NULL)
  {
    cerr << "open_rdcursor error" << endl;
    exit(-1);
  }

  while(cur->read_next(key_buf, boolvalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == BoolId &&
      be64toh(DATE(key_buf)) < end)
  {
    count++;
  }

  boolvalues->close_cursor(cur);
  callback(count);
}

void dbms::queryR4(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(double))
{
  ID(key_buf) = htobe32(AnaId);
  DATE(key_buf) = htobe64(start);

  double sum = 0;

  chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
  if(cur == NULL)
  {
    cerr << "open_rdcursor error" << endl;
    exit(-1);
  }

  while(cur->read_next(key_buf, anavalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == AnaId &&
      be64toh(DATE(key_buf)) < end)
  {
    sum+=VALUEA(anavalue_buf);
  }

  anavalues->close_cursor(cur);
  callback(sum);
}

void dbms::queryR5(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(double))
{
  ID(key_buf) = htobe32(AnaId);
  DATE(key_buf) = htobe64(start);

  uint64_t count = 0;
  double sum = 0;

  chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
  if(cur == NULL)
  {
    cerr << "open_rdcursor error" << endl;
    exit(-1);
  }

  while(cur->read_next(key_buf, anavalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == AnaId &&
      be64toh(DATE(key_buf)) < end)
  {
    sum+=VALUEA(anavalue_buf);
    count++;
  }

  anavalues->close_cursor(cur);
  callback(sum/count);
}

void dbms::queryR6(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(float, float))
{
  ID(key_buf) = htobe32(AnaId);
  DATE(key_buf) = htobe64(start);

  float min = NAN;
  float max = NAN;

  chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
  if(cur == NULL)
  {
    cerr << "open_rdcursor error" << endl;
    exit(-1);
  }

  while(cur->read_next(key_buf, anavalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == AnaId &&
      be64toh(DATE(key_buf)) < end)
  {
    if(isnan(min) || VALUEA(anavalue_buf) < min)
    {
      min = VALUEA(anavalue_buf);
    }
    if(isnan(max) || VALUEA(anavalue_buf) > max)
    {
      max = VALUEA(anavalue_buf);
    }
  }

  anavalues->close_cursor(cur);
  callback(min, max);
}

void dbms::queryR7(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(uint64_t, float))
{
  ID(key_buf) = htobe32(AnaId);
  DATE(key_buf) = htobe64(start);

  float threshold = ana_tbl[AnaId - 1].threshold2;

  chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
  if(cur == NULL)
  {
    cerr << "open_rdcursor error" << endl;
    exit(-1);
  }

  while(cur->read_next(key_buf, anavalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == AnaId &&
      be64toh(DATE(key_buf)) < end)
  {
    if(VALUEA(anavalue_buf) > threshold)
    {
      callback(be64toh(DATE(key_buf)), VALUEA(anavalue_buf));
    }
  }

  anavalues->close_cursor(cur);
}

void dbms::queryR8(uint32_t AnaId, uint64_t start,
    uint64_t end, float Threshold, void (*callback)
    (uint64_t, float))
{
  ID(key_buf) = htobe32(AnaId);
  DATE(key_buf) = htobe64(start);

  chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
  if(cur == NULL)
  {
    cerr << "open_rdcursor error" << endl;
    exit(-1);
  }

  while(cur->read_next(key_buf, anavalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == AnaId &&
      be64toh(DATE(key_buf)) < end)
  {
    if(VALUEA(anavalue_buf) > Threshold)
    {
      callback(be64toh(DATE(key_buf)), VALUEA(anavalue_buf));
    }
  }

  anavalues->close_cursor(cur);
}

void dbms::queryR9(uint64_t start, uint64_t end,
    void (*callback)(const char *, uint64_t))
{
  uint64_t counts[CFG_ANA_TAGS];

  for(unsigned int i = 1; i <= CFG_ANA_TAGS; i++)
  {
    counts[i - 1] = 0;
    float threshold1 = ana_tbl[i - 1].threshold1;
    float threshold2 = ana_tbl[i - 1].threshold2;

    ID(key_buf) = htobe32(i);
    DATE(key_buf) = htobe64(start);

    chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
    if(cur == NULL)
    {
      cerr << "open_rdcursor error" << endl;
      exit(-1);
    }
  
    while(cur->read_next(key_buf, anavalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == i &&
      be64toh(DATE(key_buf)) < end)
    {
      if(VALUEA(anavalue_buf) > threshold2 ||
          VALUEA(anavalue_buf) < threshold1)
      {
        counts[i - 1]++;
      }
    }

    anavalues->close_cursor(cur);
  }

  unsigned int max_id = 0;
  uint64_t max_count = 0;

  for(unsigned int i = 0; i < CFG_ANA_TAGS; i++)
  {
    if(counts[i] >= max_count)
    {
      max_count = counts[i];
      max_id = i;
    }
  }

  callback(ana_tbl[max_id].label, max_count);
}

void dbms::queryR10(uint64_t start, uint64_t end,
    void (*callback)(uint32_t, uint64_t))
{
  uint64_t counts[CFG_ANA_TAGS];

  for(unsigned int i = 1; i <= CFG_ANA_TAGS; i++)
  {
    counts[i - 1] = 0;
    unsigned int period = ana_tbl[i - 1].period * 1000;
    uint64_t prevdate;

    ID(key_buf) = htobe32(i);
    DATE(key_buf) = htobe64(start);

    chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);
    if(cur == NULL)
    {
      cerr << "open_rdcursor error" << endl;
      exit(-1);
    }

    cur->read_next(key_buf, anavalue_buf);
    prevdate = be64toh(DATE(key_buf));

    while(cur->read_next(key_buf, anavalue_buf) != CHR_EOF &&
      be32toh(ID(key_buf)) == i &&
      prevdate < end)
    {
      if((be64toh(DATE(key_buf)) - prevdate) > period)
      {
        counts[i - 1]++;
      }
      prevdate = be64toh(DATE(key_buf));
    }

    anavalues->close_cursor(cur);
  }
  
  unsigned int max_id = 0;
  uint64_t max_count = 0;

  for(unsigned int i = 0; i < CFG_ANA_TAGS; i++)
  {
    if(counts[i] >= max_count)
    {
      max_count = counts[i];
      max_id = i;
    }
  }
  callback(max_id + 1, max_count);
}

void dbms::queryR11_1(void (*callback)(uint32_t, float))
{
  for(unsigned int i = 1; i <= CFG_ANA_TAGS; i++)
  {
    ID(key_buf) = htobe32(i+1);
    DATE(key_buf) = htobe64(0);

    chr_cursor * cur = anavalues->open_cursor(key_buf, CHR_KEY);

    cur->read_prev(key_buf, anavalue_buf);
    
    callback(be32toh(ID(key_buf)), VALUEA(anavalue_buf));
    
    anavalues->close_cursor(cur);
  }
}

void dbms::queryR11_2(void (*callback)(uint32_t, bool))
{
  for(unsigned int i = 1; i <= CFG_TOR_TAGS; i++)
  {
    ID(key_buf) = htobe32(i+1);
    DATE(key_buf) = htobe64(0);

    chr_cursor * cur = boolvalues->open_cursor(key_buf, CHR_KEY);

    cur->read_prev(key_buf, boolvalue_buf);

    callback(be32toh(ID(key_buf)), VALUEB(boolvalue_buf));

    boolvalues->close_cursor(cur);
  }
}

