/*
 * dbms.cpp
 *
 *  Created on: 2 sept. 2009
 *      Author: chardin-bri
 */

#include "dbms.h"
#include "config.h"

#include <iostream>
#include <sstream>
#include <cstdlib>
#include <math.h>
#include <string.h>

#include <my_global.h>
#include <mysql.h>

#define CREATE_ANADESC     "CREATE TABLE ana_desc (id INT, label CHAR(40), creation_date INT, destruction_date INT, unit INT, intervl INT, threshold_low FLOAT, threshold_high FLOAT, PRIMARY KEY (id)) ENGINE = INNODB"
#define CREATE_ANAVALUES   "CREATE TABLE ana_val (id INT NOT NULL, date BIGINT UNSIGNED NOT NULL, value FLOAT NOT NULL, quality TINYINT UNSIGNED NOT NULL, PRIMARY KEY(id, date)) ENGINE = INNODB"
#define CREATE_BOOLDESC    "CREATE TABLE bool_desc (id INT, label CHAR(40), creation_date INT, destruction_date INT, label_0 CHAR(60), label_1 CHAR(60), PRIMARY KEY (id)) ENGINE = INNODB"
#define CREATE_BOOLVALUES  "CREATE TABLE bool_val (id INT NOT NULL, date BIGINT UNSIGNED NOT NULL, value BOOLEAN, quality TINYINT UNSIGNED NOT NULL, PRIMARY KEY(id, date)) ENGINE = INNODB"

#define STMT_R0_1 "INSERT INTO ana_val VALUES (?,?,?,?)"
#define STMT_R0_2 "INSERT INTO bool_val VALUES (?,?,?,?)"
#define STMT_R1_1 "UPDATE ana_val SET value=?, quality=(quality | 128) WHERE id=? AND date=?"
#define STMT_R1_2 "UPDATE bool_val SET value=?, quality=(quality | 128) WHERE id=? AND date=?"

using namespace std;

static MYSQL mysql;
static MYSQL_STMT * R0_1_stmt;
static MYSQL_BIND R0_1_bind[4];
static MYSQL_STMT * R0_2_stmt;
static MYSQL_BIND R0_2_bind[4];
static MYSQL_STMT * R1_1_stmt;
static MYSQL_BIND R1_1_bind[3];
static MYSQL_STMT * R1_2_stmt;
static MYSQL_BIND R1_2_bind[3];

static unsigned int stanaid;
static unsigned int stboolid;
static uint64_t stdate;
static float stana_value;
static int stbool_value;
static unsigned char stquality;

#define MYSQL_DATABASE  "benchmark"
#define MYSQL_USER      "brice"
#define MYSQL_SOCKET    "/tmp/mysql.sock"

const char * dbms::name()
{
  return "MySQL";
}

const char * dbms::version()
{
  return "5.5.29";
}

const char * dbms::filesystem()
{
  return "etx4";
}

void dbms::print_options()
{
  cout << "wait_timeout=360000" << endl;
  cout << "innodb_flush_log_at_trx_commit=0" << endl;
  cout << "innodb_support_xa=0" << endl;
  cout << "innodb_doublewrite=0" << endl;
  cout << "CLIENT_MULTI_RESULTS" << endl;
  cout << "datadir=/tmp/mysqldatadir" << endl;
  cout << "innodb_data_file_path=/tmp/mysqldatadir" << endl;
}

dbms::dbms()
{
  /* init mysql connection */
  if(mysql_init(&mysql)==NULL)
  {
    cerr << "Failed to initate MySQL connection" << endl;
    exit(-1);
  }
  if (mysql_real_connect(&mysql,NULL,MYSQL_USER,"",MYSQL_DATABASE,0,MYSQL_SOCKET,CLIENT_MULTI_RESULTS) == NULL)
  {
    cerr << "Failed to connect to database: Error: " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  
  /* manage tables */
  if(!CFG_SELECT_ONLY)
  {
    /* drop tables */
    if (mysql_query(&mysql, "DROP TABLE IF EXISTS ana_desc,bool_desc,ana_val,bool_val") != 0)
    {
      cerr << "DROP TABLE failed "<< mysql_error(&mysql) << endl;
      exit(-1);
    }

    /* create tables */
    if (mysql_query(&mysql, CREATE_ANADESC)!=0)
    {
      cerr << "CREATE TABLE ana_desc failed "<< mysql_error(&mysql) << endl;
      exit(-1);
    }
    if (mysql_query(&mysql, CREATE_BOOLDESC)!=0)
    {
      cerr << "CREATE TABLE bool_desc failed "<< mysql_error(&mysql) << endl;
      exit(-1);
    }
    if (mysql_query(&mysql, CREATE_ANAVALUES)!=0)
    {
      cerr << "CREATE TABLE ana_val failed "<< mysql_error(&mysql) << endl;
      exit(-1);
    }
    if (mysql_query(&mysql, CREATE_BOOLVALUES)!=0)
    {
      cerr << "CREATE TABLE bool_val failed "<< mysql_error(&mysql) << endl;
      exit(-1);
    }
  }

  /* R0.1 prepared statement */
  R0_1_stmt = mysql_stmt_init(&mysql);
  if (!R0_1_stmt)
  {
    cerr << "mysql_stmt_init(), out of memory" << endl;
    exit(-1);
  }
  if (mysql_stmt_prepare(R0_1_stmt, STMT_R0_1, strlen(STMT_R0_1)))
  {
    cerr << "mysql_stmt_prepare() failed" << mysql_stmt_error(R0_1_stmt) << endl;
    exit(-1);
  }
  memset(R0_1_bind, 0, sizeof(R0_1_bind));

  R0_1_bind[0].buffer_type = MYSQL_TYPE_LONG;
  R0_1_bind[0].buffer = (char *)&(stanaid);
  R0_1_bind[0].is_null= 0;
  R0_1_bind[0].length = 0;
  R0_1_bind[1].buffer_type = MYSQL_TYPE_LONGLONG;
  R0_1_bind[1].buffer = (char *)&(stdate);
  R0_1_bind[1].is_null= 0;
  R0_1_bind[1].length = 0;
  R0_1_bind[2].buffer_type = MYSQL_TYPE_FLOAT;
  R0_1_bind[2].buffer = (char *)&(stana_value);
  R0_1_bind[2].is_null= 0;
  R0_1_bind[2].length = 0;
  R0_1_bind[3].buffer_type = MYSQL_TYPE_LONG;
  R0_1_bind[3].buffer = (char *)&(stquality);
  R0_1_bind[3].is_null= 0;
  R0_1_bind[3].length = 0;

  /* Bind the buffers */
  if (mysql_stmt_bind_param(R0_1_stmt, R0_1_bind))
  {
    cerr << "mysql_stmt_bind_param() failed" << mysql_stmt_error(R0_1_stmt) << endl;
    exit(-1);
  }
  
  /* R0.2 prepared statement */
  R0_2_stmt = mysql_stmt_init(&mysql);
  if (!R0_2_stmt)
  {
    cerr << "mysql_stmt_init(), out of memory" << endl;
    exit(-1);
  }
  if (mysql_stmt_prepare(R0_2_stmt, STMT_R0_2, strlen(STMT_R0_2)))
  {
    cerr << "mysql_stmt_prepare() failed" << mysql_stmt_error(R0_2_stmt) << endl;
    exit(-1);
  }
  memset(R0_2_bind, 0, sizeof(R0_2_bind));

  R0_2_bind[0].buffer_type = MYSQL_TYPE_LONG;
  R0_2_bind[0].buffer = (char *)&(stboolid);
  R0_2_bind[0].is_null= 0;
  R0_2_bind[0].length = 0;
  R0_2_bind[1].buffer_type = MYSQL_TYPE_LONGLONG;
  R0_2_bind[1].buffer = (char *)&(stdate);
  R0_2_bind[1].is_null= 0;
  R0_2_bind[1].length = 0;
  R0_2_bind[2].buffer_type = MYSQL_TYPE_TINY;
  R0_2_bind[2].buffer = (char *)&(stbool_value);
  R0_2_bind[2].is_null= 0;
  R0_2_bind[2].length = 0;
  R0_2_bind[3].buffer_type = MYSQL_TYPE_LONG;
  R0_2_bind[3].buffer = (char *)&(stquality);
  R0_2_bind[3].is_null= 0;
  R0_2_bind[3].length = 0;

  /* Bind the buffers */
  if (mysql_stmt_bind_param(R0_2_stmt, R0_2_bind))
  {
    cerr << "mysql_stmt_bind_param() failed" << mysql_stmt_error(R0_2_stmt) << endl;
    exit(-1);
  }
  
  /* R1.1 prepared statement */
  R1_1_stmt = mysql_stmt_init(&mysql);
  if (!R1_1_stmt)
  {
    cerr << "mysql_stmt_init(), out of memory" << endl;
    exit(-1);
  }
  if (mysql_stmt_prepare(R1_1_stmt, STMT_R1_1, strlen(STMT_R1_1)))
  {
    cerr << "mysql_stmt_prepare() failed" << mysql_stmt_error(R1_1_stmt) << endl;
    exit(-1);
  }
  memset(R1_1_bind, 0, sizeof(R1_1_bind));

  R1_1_bind[0].buffer_type = MYSQL_TYPE_FLOAT;
  R1_1_bind[0].buffer = (char *)&(stana_value);
  R1_1_bind[0].is_null= 0;
  R1_1_bind[0].length = 0;
  R1_1_bind[1].buffer_type = MYSQL_TYPE_LONG;
  R1_1_bind[1].buffer = (char *)&(stanaid);
  R1_1_bind[1].is_null= 0;
  R1_1_bind[1].length = 0;
  R1_1_bind[2].buffer_type = MYSQL_TYPE_LONGLONG;
  R1_1_bind[2].buffer = (char *)&(stdate);
  R1_1_bind[2].is_null= 0;
  R1_1_bind[2].length = 0;

  /* Bind the buffers */
  if (mysql_stmt_bind_param(R1_1_stmt, R1_1_bind))
  {
    cerr << "mysql_stmt_bind_param() failed" << mysql_stmt_error(R1_1_stmt) << endl;
    exit(-1);
  }
  
    /* R0.1 prepared statement */
  R1_2_stmt = mysql_stmt_init(&mysql);
  if (!R1_2_stmt)
  {
    cerr << "mysql_stmt_init(), out of memory" << endl;
    exit(-1);
  }
  if (mysql_stmt_prepare(R1_2_stmt, STMT_R1_2, strlen(STMT_R1_2)))
  {
    cerr << "mysql_stmt_prepare() failed" << mysql_stmt_error(R1_2_stmt) << endl;
    exit(-1);
  }
  memset(R1_2_bind, 0, sizeof(R1_2_bind));

  R1_2_bind[0].buffer_type = MYSQL_TYPE_TINY;
  R1_2_bind[0].buffer = (char *)&(stbool_value);
  R1_2_bind[0].is_null= 0;
  R1_2_bind[0].length = 0;
  R1_2_bind[1].buffer_type = MYSQL_TYPE_LONG;
  R1_2_bind[1].buffer = (char *)&(stboolid);
  R1_2_bind[1].is_null= 0;
  R1_2_bind[1].length = 0;
  R1_2_bind[2].buffer_type = MYSQL_TYPE_LONGLONG;
  R1_2_bind[2].buffer = (char *)&(stdate);
  R1_2_bind[2].is_null= 0;
  R1_2_bind[2].length = 0;

  /* Bind the buffers */
  if (mysql_stmt_bind_param(R1_2_stmt, R1_2_bind))
  {
    cerr << "mysql_stmt_bind_param() failed" << mysql_stmt_error(R1_2_stmt) << endl;
    exit(-1);
  }
}

dbms::~dbms()
{
  if (mysql_stmt_close(R0_1_stmt))
  {
    cerr << "failed while closing the statement" << mysql_stmt_error(R0_1_stmt) << endl;
    exit(-1);
  }
  if (mysql_stmt_close(R0_2_stmt))
  {
    cerr << "failed while closing the statement" << mysql_stmt_error(R0_2_stmt) << endl;
    exit(-1);
  }
  if (mysql_stmt_close(R1_1_stmt))
  {
    cerr << "failed while closing the statement" << mysql_stmt_error(R1_1_stmt) << endl;
    exit(-1);
  }
  if (mysql_stmt_close(R1_2_stmt))
  {
    cerr << "failed while closing the statement" << mysql_stmt_error(R1_2_stmt) << endl;
    exit(-1);
  }

  mysql_close(&mysql);
}

void dbms::insertAna(uint32_t anaid, char * label, unsigned int creationdate, unsigned int * destructiondate, unsigned int unit, unsigned int period, float threshold1, float threshold2)
{
  if(!CFG_SELECT_ONLY)
  {
    ostringstream query;
    query << "INSERT INTO ana_desc VALUES (" << anaid << ", \"" << label
      << "\", " << creationdate << ", ";
    
    if(destructiondate == NULL)
    {
      query << "NULL";
    }
    else
    {
      query << *destructiondate;
    }
    
    query << ", " << unit << ", " << period << ", " << threshold1
      << ", " << threshold2 << ")";
    
    if(mysql_query(&mysql, query.str().c_str()))
    {
      cerr << "insertAna error " << mysql_error(&mysql) << endl;
      exit(-1);
    }
  }
}

void dbms::insertBool(uint32_t boolid, char * label, unsigned int creationdate, unsigned int * destructiondate, char * label0, char * label1)
{
  if(!CFG_SELECT_ONLY)
  {
    ostringstream query;
    query << "INSERT INTO bool_desc VALUES (" << boolid << ", \"" << label
      << "\", " << creationdate << ", ";
    
    if(destructiondate == NULL)
    {
      query << "NULL";
    }
    else
    {
      query << *destructiondate;
    }
    
    query << ", \"" << label0 << "\", \"" << label1 << "\")";
    
    if(mysql_query(&mysql, query.str().c_str()))
    {
      cerr << "insertBool error " << mysql_error(&mysql) << endl;
      exit(-1);
    }
  }
}

void dbms::flushR0_1()
{

}

void dbms::flushR0_2()
{

}

void dbms::queryR0_1(uint32_t anaid, uint64_t date, float ana_value, unsigned char quality)
{
  if(!CFG_SELECT_ONLY)
  {
    stanaid = anaid;
    stdate = date;
    stana_value = ana_value;
    stquality = quality;
    
    if(mysql_stmt_execute(R0_1_stmt))
    {
      cerr << "mysql_stmt_execute() failed " << mysql_stmt_error(R0_1_stmt) << endl;
      exit(-1);
    }
  }
}

void dbms::queryR0_2(uint32_t boolid, uint64_t date,
    bool bool_value, unsigned char quality)
{
  if(!CFG_SELECT_ONLY)
  {
    stboolid = boolid;
    stdate = date;
    if(bool_value)
    {
      stbool_value = 1;
    }
    else
    {
      stbool_value = 0;
    }
    stquality = quality;
    
    if(mysql_stmt_execute(R0_2_stmt))
    {
      cerr << "mysql_stmt_execute() failed " << mysql_stmt_error(R0_2_stmt) << endl;
      exit(-1);
    }
  }
}

void dbms::queryR1_1(float value, uint32_t anaid, uint64_t date)
{
  if(!CFG_SELECT_ONLY)
  {
    stana_value = value;
    stanaid = anaid;
    stdate = date;
    
    if(mysql_stmt_execute(R1_1_stmt))
    {
      cerr << "mysql_stmt_execute() failed " << mysql_stmt_error(R1_1_stmt) << endl;
      exit(-1);
    }
  }
}

void dbms::queryR1_2(bool value, uint32_t boolid,
    uint64_t date)
{
  if(!CFG_SELECT_ONLY)
  {
    if(value)
    {
      stbool_value = 1;
    }
    else
    {
      stbool_value = 0;
    }
    stboolid = boolid;
    stdate = date;
    
    if(mysql_stmt_execute(R1_2_stmt))
    {
      cerr << "mysql_stmt_execute() failed " << mysql_stmt_error(R1_2_stmt) << endl;
      exit(-1);
    }
  }
}

void dbms::queryR2_1(uint32_t anaid, uint64_t start,
    uint64_t end, void (*callback)
    (unsigned int, uint64_t, float, unsigned char))
{
  ostringstream query;
  query << "SELECT * FROM ana_val WHERE id=" << anaid << " AND date BETWEEN " << start << " AND " << end;
  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR2_1 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    /* should couvert these values */
    callback(0, 0, 0, 0);
  }
  mysql_free_result(result);
}

void dbms::queryR2_2(uint32_t boolid, uint64_t start,
    uint64_t end, void (*callback)
    (unsigned int, uint64_t, bool, unsigned char))
{
  ostringstream query;
  query << "SELECT * FROM bool_val WHERE id=" << boolid << " AND date BETWEEN " << start << " AND " << end;
  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR2_2 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    /* should couvert these values */
    callback(0, 0, 0, 0);
  }
  mysql_free_result(result);
}

void dbms::queryR3_1(uint32_t anaid, uint64_t start,
    uint64_t end, void (*callback)(uint64_t))
{
  ostringstream query;
  query << "SELECT count(*) FROM ana_val WHERE id=" << anaid << " AND date BETWEEN " << start << " AND " << end;

  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR3_1 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    callback(atoi(row[0]));
  }
  mysql_free_result(result);
}

void dbms::queryR3_2(uint32_t boolid, uint64_t start,
    uint64_t end, void (*callback)(uint64_t))
{
  ostringstream query;
  query << "SELECT count(*) FROM bool_val WHERE id=" << boolid << " AND date BETWEEN " << start << " AND " << end;
  
  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR3_2 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    callback(atoi(row[0]));
  }
  mysql_free_result(result);
}

void dbms::queryR4(uint32_t anaid, uint64_t start,
    uint64_t end, void (*callback)(double))
{
  ostringstream query;
  query << "SELECT sum(value) FROM ana_val WHERE id=" << anaid << " AND date BETWEEN " << start << " AND " << end;

  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR4 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    if(row[0] != NULL)
    {
      callback(atof(row[0]));
    } else {
      callback(NAN);
    }
  }
  mysql_free_result(result);
}

void dbms::queryR5(uint32_t anaid, uint64_t start,
    uint64_t end, void (*callback)(double))
{
  ostringstream query;
  query << "SELECT avg(value) FROM ana_val WHERE id=" << anaid << " AND date BETWEEN " << start << " AND " << end;

  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR5 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    if(row[0] != NULL)
    {
      callback(atof(row[0]));
    } else {
      callback(NAN);
    }
  }
  mysql_free_result(result);
}

void dbms::queryR6(uint32_t anaid, uint64_t start,
    uint64_t end, void (*callback)(float, float)) {
  ostringstream query;
  query << "SELECT min(value), max(value) FROM ana_val WHERE id=" << anaid << " AND date BETWEEN " << start << " AND " << end;

  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR6 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    if(row[0] != NULL)
    {
      callback(atof(row[0]), atof(row[1]));
    } else {
      callback(NAN, NAN);
    }
  }
  mysql_free_result(result);
}

void dbms::queryR7(uint32_t anaid, uint64_t start,
    uint64_t end, void (*callback)(uint64_t, float))
{
  ostringstream query;
  query << "SELECT date, value FROM ana_desc, ana_val WHERE ana_desc.id = ana_val.id AND ana_val.id=" << anaid << " AND date BETWEEN " << start << " AND " << end << " AND value>ana_desc.threshold_high";

  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR7 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    /* should couvert these values */
    callback(0, 0);
  }
  mysql_free_result(result);
}

void dbms::queryR8(uint32_t anaid, uint64_t start,
    uint64_t end, float Threshold, void (*callback)
    (uint64_t, float))
{
  ostringstream query;
  query << "SELECT date, value FROM ana_val WHERE id=" << anaid << " AND date BETWEEN " << start << " AND " << end << " AND value>" << Threshold;

  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR8 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    /* should couvert these values */
    callback(0, 0);
  }
  mysql_free_result(result);
}

/*
typedef struct tmpR9_st
{
  char label[80];
  uint32_t anaid;
  float threshold1;
  float threshold2;
} tmpR9_t;
*/

void dbms::queryR9(uint64_t start, uint64_t end,
    void (*callback)(const char *, uint64_t))
{
  /*
   query << "SELECT label, count(*) as count FROM ana_desc STRAIGHT JOIN ana_val FORCE INDEX (PRIMARY) ON ana_desc.id=ana_val.id WHERE date BETWEEN " << start << " AND " << end << " AND (value>threshold_high OR value<threshold_low) GROUP BY ana_desc.id ORDER BY count DESC LIMIT 1";
  */
  ostringstream query;
  query << "CALL R9(" << start << ", " << end << ")";
  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR9 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    if(row[0] != NULL)
    {
      callback(row[0], atoi(row[1]));
    } else {
      callback("", 0);
    }
  }
  mysql_free_result(result);
  while(mysql_next_result(&mysql) > 0)
  {
    result = mysql_use_result(&mysql);
    while(mysql_fetch_row(result) != NULL);
    mysql_free_result(result);
  }

  /*
  char label[80] = "\0";
  uint64_t count = 0;

  ostringstream query;
  query << "SELECT label, id, threshold_low, threshold_high FROM ana_desc";
  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR9 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  unsigned int ids = mysql_num_rows(result);
  tmpR9_t * tmpres = (tmpR9_t *)malloc(ids * sizeof(tmpR9_t));
  for(int i=0; (row = mysql_fetch_row(result)) != NULL; i++)
  {
    strcpy(tmpres[i].label, row[0]);
    tmpres[i].anaid = atoi(row[1]);
    tmpres[i].threshold1 = atof(row[2]);
    tmpres[i].threshold2 = atof(row[3]);
  }
  mysql_free_result(result);

  for(unsigned int i=0; i < ids ; i++)
  {
    ostringstream subquery;
    subquery << "SELECT count(*) FROM ana_val WHERE id = "
      << tmpres[i].anaid << " AND (value>" << tmpres[i].threshold2
      << " OR value<" << tmpres[i].threshold1 << ") AND date BETWEEN "
      << start << " AND " << end;
    if(mysql_query(&mysql, subquery.str().c_str()))
    {
      cerr << "subqueryR9 (" << tmpres[i].anaid <<") error "
        << mysql_error(&mysql) << endl;
      exit(-1);
    }
    MYSQL_RES * subresult = mysql_use_result(&mysql);
    MYSQL_ROW subrow;
    while((subrow = mysql_fetch_row(subresult)) != NULL)
    {
      if((unsigned int)atoi(subrow[0]) > count)
      {
        count = atoi(subrow[0]);
        strcpy(label, tmpres[i].label);
      }
    }
    mysql_free_result(subresult);
  }
  callback(label, count);
  free(tmpres);
  */
}

void dbms::queryR10(uint64_t start, uint64_t end,
    void (*callback)(unsigned int, uint64_t))
{
  ostringstream query;
  /*
  query << "SELECT Diff.id, count(*) as c, MYSQL FROM ana_desc, (SELECT D1.id, D1.date, min(D2.date-D1.date) as diff_ms FROM ana_val D1, ana_val D2 WHERE D2.date>D1.date AND D1.id=D2.id AND D1.date BETWEEN " << start << " AND " << end << " GROUP BY D1.id, D1.date) as Diff WHERE Diff.id = ana_desc.id AND Diff.diff_ms>ana_desc.MYSQL GROUP BY Diff.id ORDER BY c DESC LIMIT 1";
  */
  query << "CALL R10(" << start << ", " << end << ")";

  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR10 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    if(row[0] != NULL) {
      callback(atoi(row[0]), atoi(row[1]));
    } else {
      callback(0, 0);
    }
  }
  mysql_free_result(result);
  while(mysql_next_result(&mysql) > 0)
  {
    result = mysql_use_result(&mysql);
    while(mysql_fetch_row(result) != NULL);
    mysql_free_result(result);
  }
}

void dbms::queryR11_1(void (*callback)(unsigned int, float))
{
  ostringstream query;
  query << "SELECT ana_val.id, value FROM ana_val, (SELECT id, max(date) as maxdate FROM ana_val GROUP BY id) as maxdates WHERE date=maxdates.maxdate AND ana_val.id = maxdates.id ORDER BY ana_val.id";

  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR11.1 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    callback(atoi(row[0]), atof(row[1]));
  }
  mysql_free_result(result);
}

void dbms::queryR11_2(void (*callback)(unsigned int, bool))
{
  ostringstream query;
  query << "SELECT bool_val.id, value FROM bool_val FORCE INDEX (PRIMARY), (SELECT id, max(date) as maxdate FROM bool_val FORCE INDEX (PRIMARY) GROUP BY id) as maxdates WHERE date=maxdates.maxdate AND bool_val.id = maxdates.id ORDER BY bool_val.id";
  
  if(mysql_query(&mysql, query.str().c_str()))
  {
    cerr << "queryR11.2 error " << mysql_error(&mysql) << endl;
    exit(-1);
  }
  MYSQL_RES * result = mysql_use_result(&mysql);
  MYSQL_ROW row;
  while((row = mysql_fetch_row(result)) != NULL)
  {
    callback(atoi(row[0]), atoi(row[1]));
  }
  mysql_free_result(result);
}

